/*
 * @copyright(disclaimer) 
 * Licensed Materials - Property of IBM
 * IBM  Enterprise Information Portal  for Multiplatforms V8.1  (program number 5724-B43)
 * (c ) Copyright IBM Corp. 1998,  2002. All Rights Reserved.
 * 
 * US Government Users Restricted Rights                                  
 *  Use, duplication or disclosure restricted by GSA ADP Schedule          
 *  Contract with IBM Corp.     
 * 
 * DISCLAIMER OF WARRANTIES :                                             
 *                                                                        
 * Permission is granted to copy and modify this  Sample code, and to           
 * distribute modified versions provided that both the copyright        
 * notice,- and this permission notice and warranty disclaimer  appear in all copies and modified  
 * versions. 
 * 
 * THIS SAMPLE CODE IS LICENSED TO YOU AS-IS. 
 * IBM  AND ITS SUPPLIERS AND LICENSORS  DISCLAIM     
 * ALL WARRANTIES, EITHER EXPRESS OR IMPLIED, IN 
 * SUCH SAMPLE CODE, INCLUDING THE WARRANTY OF 
 * NON-INFRINGEMENT AND THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR 
 * PURPOSE. IN NO EVENT WILL IBM OR ITS LICENSORS OR
 * SUPPLIERS BE LIABLE FOR ANY DAMAGES ARISING OUT 
 * OF THE USE OF  OR INABILITY TO USE THE SAMPLE CODE, 
 * DISTRIBUTION OF THE SAMPLE CODE, OR COMBINATION 
 * OF THE SAMPLE CODE WITH ANY OTHER CODE. IN NO 
 * EVENT SHALL IBM OR ITS LICENSORS AND SUPPLIERS BE 
 * LIABLE FOR ANY LOST REVENUE, LOST PROFITS OR DATA, 
 * OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED 
 * AND REGARDLESS OF THE THEORY OF LIABILITY,-, EVEN IF 
 * IBM OR ITS LICENSORS OR SUPPLIERS HAVE BEEN ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGES.                                     
 *                                                                  
 * @endCopyright
 */

import java.awt.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;


/**
 * Displays messages in a modal dialog.  In its simplest form the user passes
 * a title, message and width to the constructor and gets a message box with 
 * an "OK" button.  For more complicated situations the user can specify what
 * buttons are to appear and what ActionCommand they will return in addition
 * to setting the message text, title and width.  With the simple method the
 * message box is automatically displayed.  In the more complex case, the 
 * message box is not displayed until one of the "display" methods is called.
 *
 * <p>The width parameter is used to set the number of columns in the displayed
 * message.  The maximum width is 120 and the minimum width is 10.  If the
 * maximum width is larger than the current screen resolution will support then
 * the maximum width will be reduced so that the message will fit on the screen.
 *
 * <p>The message is word wrapped so that words are not split across lines.
 * Any linefeed characters (\n) are retained and cause new lines in the
 * display.
 *
 * <p>You can also register a single ActionListener when creating a CMBMessageBox.  
 * This will allow notification of the button selection to your code.
 *
 * <p>You must supply the frame of the window you wish focus to return to.
 * If you use a value of null the focus returns to the desktop.
 *
 * <p>By default, the message box operates in an asynchronous manner.  You can
 * use setSynchronousMode to change this behaviour.
 *
 * <p>An alternative to passing in a simple string as a message, is the 
 * ability to pass in a message ID and array of values (or a single value) to
 * plug into the message.  Uses the ResourceMgr class to get the message
 * based on its ID.
 *
 * @see JOptionPane#getFrameForComponent
 * @see ResourceMgr
 */
public class MessageBox implements Runnable, 
ActionListener, 
WindowListener, 
KeyListener
{
   // Do not remove - needed for copyright protection
   private static String COPYRIGHT = ClientPane.IBM_LONG_COPYRIGHT;

   private String  fieldCloseWindowCommand = "Message Closed";  // Don't Translate
   private String  fieldTitle;

   private transient ActionListener  fieldListener = null;
   private transient Frame           fieldFrame = null;
   private transient boolean         fieldFrameNotProvided = false;
   private transient JPanel          fieldButtonPanel = null;
   private transient String          fieldMessageText;  //This is the unwrapped line of text.
   private transient JDialog         fieldMessageBox;
   private transient int             fieldWidth;
   private transient String          fieldMessageID = null;
   private transient boolean         fieldSynchronousMode = false;
   private transient JButton         fieldFirstButton = null;

   private final static int   minWidth = 10;
   private static int         maxWidth = 0;



/**
 * Creates an empty undisplayed message box.
 */
   public MessageBox()
   {
      //EJMTrace.trace( EJMTrace.DEBUG, this, "Starting new CMBMessageBox");
      setLookAndFeel();
      fieldButtonPanel = new JPanel();
      fieldWidth = minWidth;
      if (maxWidth < minWidth)
      {
         setMaximumWidth();
      }
   }


/**
 * Creates an empty undisplayed message box and assigns a title, width and
 * a listener for results.
 *
 * @param newTitle            the title of the message window
 * @param newWidth            the width in columns of the message display
 * @param caller              the object listening for results
 */
   public MessageBox( String newTitle, int newWidth, ActionListener caller)
   {
      this();
      fieldListener = caller;
      setTitle( newTitle);
      setWidth( newWidth);
   }


/**
 * Displays a default message box with an "OK" button.  Uses the provided message,
 * title and width and sets a listener for results.
 *
 * @param newFrame            the parent frame for the message box
 * @param newTitle            the title of the message window
 * @param message             the text to display in the message box
 * @param newWidth            the width in columns of the message display
 * @param caller              the object listening for results
 */
   public MessageBox( Frame newFrame, String newTitle, String message, int newWidth, ActionListener caller)
   {
      this();
      fieldFrame = newFrame;
      fieldListener = caller;
      fieldMessageText = message;
      setTitle( newTitle);
      setWidth( newWidth); 
      addChoice(ResourceMgr.getString("OK"));
      display( fieldMessageText);
   }

/**
 * Displays a default message box with an "OK" button.  Uses the provided message ID,
 * list of parameters, title and width and sets a listener for results.
 *
 * @param newFrame            the parent frame for the message box
 * @param newTitle            the title of the message window
 * @param messageID           the ID of the message to be displayed
 * @param args                an array of arguments to substitue in the message
 * @param includeID           should the message ID be included in the text
 * @param newWidth            the width in columns of the message display
 * @param caller              the object listening for results
 */
   public MessageBox( Frame newFrame, String newTitle, String messageID, Object[] args, boolean includeID, int newWidth, ActionListener caller)
   {
      this();
      fieldFrame = newFrame;
      fieldListener = caller;
      fieldMessageID = messageID;

      setMessageText( messageID, args, includeID);
      setTitle( newTitle);
      setWidth( newWidth); 
      addChoice(ResourceMgr.getString("OK"));
      display( fieldMessageText);
   }


/**
 * Displays a default message box with an "OK" button.  Uses the provided message ID,
 * a sinlge replaceable parameter, title and width and sets a listener for results.
 *
 * @param newFrame            the parent frame for the message box
 * @param newTitle            the title of the message window
 * @param messageID           the ID of the message to be displayed
 * @param arg                 a single argument to substitue in the message
 * @param includeID           should the message ID be included in the text
 * @param newWidth            the width in columns of the message display
 * @param caller              the object listening for results
 */
   public MessageBox( Frame newFrame, String newTitle, String messageID, Object arg, boolean includeID, int newWidth, ActionListener caller)
   {
      this();
      fieldFrame = newFrame;
      fieldListener = caller;
      fieldMessageID = messageID;

      setMessageText( messageID, arg, includeID);
      setTitle( newTitle);
      setWidth( newWidth); 
      addChoice(ResourceMgr.getString("OK"));
      display( fieldMessageText);
   }


//---------- Runnable Implementation ---------------------
/**
 * This prevents the caller from blocking on display(), which
 * if this class is used on an awt event thread could
 * cause a deadlock.
 */
   public void run()
   {
      if (fieldFirstButton != null)
      {
         fieldFirstButton.requestFocus();
      }
      fieldMessageBox.setVisible( true);
   }


//---------- ActionListener Implementation ---------------
/**
 * Passes events from the message box to the registered listener
 */
   public void actionPerformed( ActionEvent evt)
   {
      String command = evt.getActionCommand();
      fieldMessageBox.setVisible( false);
      fieldMessageBox.dispose();

      //EJMTrace.trace( EJMTrace.DBALL, "CMBMessageBox action selected: ", command);
      if ( fieldFrameNotProvided)
         fieldFrame.dispose();

      if ( fieldListener != null)
      {
         fieldListener.actionPerformed( evt);
      }
   }


//---------- WindowListener Implementatons ---------------
   public void windowClosing( WindowEvent evt) 
   {
      // User clicked on X or chose Close selection
      fireCloseRequested();
   }

   public void windowClosed( WindowEvent evt)
   {}

   public void windowDeiconified( WindowEvent evt)
   {}

   public void windowIconified( WindowEvent evt)
   {}

   public void windowOpened( WindowEvent evt)
   {}

   public void windowActivated( WindowEvent evt)
   {}

   public void windowDeactivated( WindowEvent evt)
   {}

//---------- KeyListener Implementation ------------------
   public void keyTyped( KeyEvent evt)
   {}

   public void keyPressed( KeyEvent evt) 
   {
      if (evt.getKeyCode() == KeyEvent.VK_ESCAPE)
      {
         fireCloseRequested();
      }

      if (evt.getKeyCode() == KeyEvent.VK_TAB)
      {
//         System.out.println( "Focus - " + fieldFrame.getFocusOwner());
      }
   }


   public void keyReleased( KeyEvent evt)
   {}

   private void fireCloseRequested()
   {
      ActionEvent event = new ActionEvent( this, ActionEvent.ACTION_PERFORMED, fieldCloseWindowCommand);
      actionPerformed( event);
   }

/**
 * Sets the listener to be notified of button clicks
 * and WindowClosing events.
 *
 * @param   caller         the listener for results from the message box
 */
   public void setActionListener( ActionListener caller) 
   {
      fieldListener = caller;
   }


/** 
 * Sets the title of the Message window.  If a null title
 * is sent then the title is blank.
 *
 * @param  newTitle        the title of the message window
 */
   public void setTitle( String newTitle)
   {
      if (newTitle != null)
      {
         fieldTitle = newTitle;
      }
      else
         fieldTitle = "";
   }


/**
 * Sets the line width of the displayed message.  A width of
 * less than 10 or greater than 120 is not allowed. If the width is
 * larger than the screen resolution will support, the maximum size
 * will be reduced so that the message will fit on the screen.
 *
 * @param  newWidth           the width of the message display in columns
 */
   public void setWidth( int newWidth)
   {
      if ( newWidth >= minWidth)
      {
         fieldWidth = newWidth;
      }
      else
         fieldWidth = minWidth;

      if ( newWidth > maxWidth)
      {
         fieldWidth = maxWidth;
      }
   }


/**
 * Sets the text of the message to be displayed.
 *
 * @param newMessage          the text to be displayed
 */
   public void setMessageText( String newMessage)
   {

      if ( newMessage == null)
      {
         newMessage = "";
      }
      fieldMessageText = newMessage;

   }


/**
 * Sets the text of the message to be displayed based on a message ID
 *
 * @param  messageID          the ID of the message to be displayed
 * @param  args               the replaceable parameters in the message
 * @param  includeID          should ID number be included in message text
 */
   public void setMessageText( String messageID, Object[] args, boolean includeID)
   {

      if ( messageID == null)
      {
         fieldMessageText = "";
         return;
      }

      if ( args == null)
      {
         if ( includeID)
         {
            fieldMessageText = ResourceMgr.getStringWithMessageNumber( messageID);
         }
         else
         {
            fieldMessageText = ResourceMgr.getString( messageID);
         }
      }
      else
      {
         if ( includeID)
         {
            fieldMessageText = ResourceMgr.getStringWithMessageNumber( messageID, args);
         }
         else
         {
            fieldMessageText = ResourceMgr.getString( messageID, args);
         }
      }

   }


/**
 * Sets the text of the message to be displayed based on a message ID
 *
 * @param  messageID          the ID of the message to be displayed
 * @param  arg                the single replaceable parameter in the message
 * @param  includeID          should ID number be included in message text
 */
   public void setMessageText( String messageID, Object arg, boolean includeID)
   {

      if ( messageID == null)
      {
         fieldMessageText = "";
         return;
      }

      if ( arg == null)
      {
         if ( includeID)
         {
            fieldMessageText = ResourceMgr.getStringWithMessageNumber( messageID);
         }
         else
         {
            fieldMessageText = ResourceMgr.getString( messageID);
         }
      }
      else
      {
         if ( includeID)
         {
            fieldMessageText = ResourceMgr.getStringWithMessageNumber( messageID, arg);
         }
         else
         {
            fieldMessageText = ResourceMgr.getString( messageID, arg);
         }
      }
   }


/**
 * If a Frame is provided then it is used to instantiate
 * the modal Dialog. Otherwise a temporary Frame is used.
 * Providing a Frame will have the effect of putting the
 * focus back on that Frame when the MessageBox is closed
 * or a button is clicked.
 */
   public void setFrame( Frame newFrame)  // Optional
   {
      fieldFrame = newFrame;
   }

/**
 * Indicates whether or not the message box is set to 
 * operate as a synchronous call or as an asynchronous call.
 *
 * @param   newMode         true for synchronous operation, false otherwise
 */
   public boolean isSynchronous( )  // Optional
   {
      return fieldSynchronousMode;
   }


/**
 * Sets the mode of operation for the message box as either
 * a synchronous call or as an asynchronous call.
 *
 * @param   newMode         true for synchronous operation, false otherwise
 */
   public void setSynchronousMode( boolean newMode)  // Optional
   {
      fieldSynchronousMode = newMode;
   }

/**
 * Sets the ActionCommand used in the ActionEvent when the
 * user attempts to close the window. The window may be
 * closed by clicking on "X", choosing Close from the
 * window menu, or pressing the Escape key. The default
 * command is "CloseRequested", which is just what a Close
 * choice button would probably have as a command.
 */
   public void setCloseWindowCommand( String command) 
   {
      fieldCloseWindowCommand = command;
   }


/**
 * The @param label will be used for the button and the
 * @param command will be returned to the listener.
 */
   public void addChoice( String label, String command)
   {
      JButton button = new JButton( label);
      button.setActionCommand( command);
      button.addActionListener( this);
      button.addKeyListener( this);
      fieldButtonPanel.add( button);

      // Remember first button so we can set focus later
      if (fieldFirstButton == null)
      {
         fieldFirstButton = button;
      }
   }

/**
 * A convenience method that assumes the command is the
 * same as the label.
 */
   public void addChoice( String label)
   {
      addChoice( label, label);
   }


/**
 * Requests that the message box be displayed with the already set parameters.
 * The message text should have already been set.  Needed if the default 
 * "OK" message box constructor is not used.
 * 
 * @see #setMessageText(String)
 *
 * @exception   CMBShouldNotOccurException  on a null message
 */
 public void display()
{
   if (fieldMessageText != null)
   {
      display( fieldMessageText);
   }
   else
      throw new ShouldNotOccurException( "Display message is null");
}


/**
 * Displays the message box using the provided text.
 * One of the "display" methods is needed to actually display the
 * message box if the default "OK" message box constructor is
 * not used.
 *
 * @param  message      the text to be displayed
 *
 * @exception   CMBShouldNotOccurException  on a null message
 */
   public void display( String message) 
   {
      if ( message == null)
      {         
         throw new ShouldNotOccurException( "Display message is null");
      }

      // Format the message for the requested width
      message = wrap2( message, fieldWidth, null);

      // Adjust for overly wide box widths
      if ( message.length() < (fieldWidth)) 
      {
         setWidth( message.length() + 1);
      }

      if ( fieldFrame == null)
      {
         fieldFrame = new Frame();
         fieldFrameNotProvided = true;
      }
      else
      {
         fieldFrameNotProvided = false;
      }
      fieldMessageBox = new JDialog( fieldFrame, true); // Modal
      fieldMessageBox.addWindowListener( this);
      fieldMessageBox.addKeyListener( this);
      fieldMessageBox.setTitle( fieldTitle);
      fieldMessageBox.getContentPane().setLayout(new BorderLayout(10, 15));
      fieldMessageBox.setResizable( false);

      JTextArea textArea = new JNFTextArea( message, 1, fieldWidth + 1);
      textArea.setOpaque( false);
      textArea.setEditable( false);
      textArea.setBorder( new EmptyBorder( 1, 1, 1, 1));
//      textArea.setLineWrap( true);
      JPanel messagePanel = new JPanel( new FlowLayout());
      messagePanel.add( textArea);

      fieldMessageBox.getContentPane().add( messagePanel, BorderLayout.CENTER);
      fieldMessageBox.getContentPane().add( fieldButtonPanel, BorderLayout.SOUTH);
      fieldMessageBox.getContentPane().add( new JPanel(), BorderLayout.NORTH);
      fieldMessageBox.getContentPane().add( new JPanel(), BorderLayout.WEST);
      fieldMessageBox.getContentPane().add( new JPanel(), BorderLayout.EAST);

      fieldMessageBox.pack();
      enforceMinimumSize( fieldMessageBox, 200, 100);
      centerWindow( fieldMessageBox);
//    Toolkit.getDefaultToolkit().beep();

      traceLog( fieldMessageID, fieldMessageText);

      if ( fieldSynchronousMode)
      {
         if (fieldFirstButton != null)
         {
            fieldFirstButton.requestFocus();
         }

         fieldMessageBox.setVisible( true);
      }
      else
      {
         // Start a new thread to show the dialog
         Thread thread = new Thread( this);
         thread.start();
      }
   }



/**
 * Sets the maximum number of columns based on screen resolution.
 * Does not adjust with runtime changes in screen resolution.
 *
 */
   private void setMaximumWidth( )
   {
      Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();

      JTextArea temp = new JTextArea();
      // This is the way JTextArea actually calculates prefered size from column width (swing 1.0.2)
      FontMetrics metrics = temp.getFontMetrics( temp.getFont());
      int columnWidth = metrics.charWidth('m');
      maxWidth = (screenDim.width / columnWidth) - 8;
      if (maxWidth > 120)
         maxWidth = 120;
   }



/**
 * Centers the message box on the screen
 *
 * @param  win          the window to reposition
 */
   private void centerWindow( Window win)
   {
      Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
      // If larger than screen, reduce window width or height
      if ( screenDim.width < win.getSize().width)
      {
         win.setSize( screenDim.width, win.getSize().height);
      }

      if (screenDim.height < win.getSize().height)
      {
         win.setSize( win.getSize().width, screenDim.height);
      }

      // Center Frame, Dialogue or Window on screen
      int x = (screenDim.width - win.getSize().width) / 2;
      int y = (screenDim.height - win.getSize().height) / 2;
      win.setLocation(x, y);
   }


/*
 * Sets the message box to have a minimum size.
 *
 * @param   comp           the message box
 * @param   minWidth       the minimum width
 * @param   minHeight      the minimum height
 */
   private void enforceMinimumSize( Component comp, int minWidth, int minHeight) 
   {
      if ( comp.getSize().width < minWidth)
      {
         comp.setSize( minWidth, comp.getSize().height);
      }

      if (comp.getSize().height < minHeight)
      {
         comp.setSize( comp.getSize().width, minHeight);
      }
   }


/**
 * Sets the "look and feel" to the System defined "look and feel".
 */
   private void setLookAndFeel()
   {
      try
      {
         UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
      }
      catch (java.lang.IllegalAccessException exc)
      {
         //EJMTrace.trace(EJMTrace.INFO,this,exc.toString() + ":" + exc.getMessage());
      }
      catch (java.lang.InstantiationException   exc)
      {
         //EJMTrace.trace(EJMTrace.INFO,this,exc.toString() + ":" + exc.getMessage());
      }
      catch (java.lang.ClassNotFoundException  exc)
      {
         //EJMTrace.trace(EJMTrace.INFO,this,exc.toString() + ":" + exc.getMessage());
      }
      catch (javax.swing.UnsupportedLookAndFeelException exc)
      {
         //EJMTrace.trace(EJMTrace.INFO,this,exc.toString() + ":" + exc.getMessage());
      }
   }


/**
 *
 * This method takes a string and wraps it to a line length of no more than
 * wrap_length.  If prepend is not null, each resulting line will be prefixed
 * with the prepend string.  In that case, resultant line length will be no
 * more than wrap_length + prepend.length()
 *
 * @param inString         the string to wrap
 * @param wrap_length      the desired number of columns in each line 
 * @param prepend          the string to prepend to each line
 */
   private String wrap( String inString, int wrap_length, String prepend)
   {
      char[] charAry;

      int p;              
      int p2;
      int offset = 0;
      int marker;

      StringBuffer result = new StringBuffer();

      if ( inString == null) 
      {
         throw new ShouldNotOccurException( "MessageBox - null message");
      }

      if ( wrap_length < 0)
      {
         throw new ShouldNotOccurException( "MessageBox - width too small: " + wrap_length);
      }


      if (prepend != null)
      {
         result.append(prepend);
      }

      //EJMTrace.trace( EJMTrace.DEBUG, this, "String size = " + inString.length());
 
      charAry = inString.toCharArray();

      p = marker = 0;

      // each time through the loop, p starts out pointing to the same char as marker

      while (marker < charAry.length)
      {
         while (p < charAry.length && (charAry[ p] != '\n') && ((p - marker) < wrap_length))
         {
            p++;
         }

         if (p == charAry.length)
         {
            //EJMTrace.trace( EJMTrace.DEBUG, this, "Completed parsing message");

            result.append( inString.substring( marker, p));
            return result.toString();
         }

         //EJMTrace.trace( EJMTrace.DEBUG, this, "Step 1: p = " + p + ", marker = " + marker);
         if (charAry[p] == '\n')
         {
            /* We've got a newline.  This newline is bound to have terminated
               the while loop above.  Step p back one character so that the 
               isspace(*p) check below will detect that it hit the \n, and
               will do the right thing. */

            result.append( inString.substring( marker, p+1));

            if (prepend != null)
            {
               result.append(prepend);
            }

            //EJMTrace.trace( EJMTrace.DEBUG, this, "found natural newline.. current result = " + result.toString());

            p = marker = p+1;

            continue;
         }

         //EJMTrace.trace( EJMTrace.DEBUG, this, "Step 2: hit wrap length, back searching for newline");

         p2 = p-1;

         /* We've either hit the end of the string, or we've gotten past the 
            wrap_length.  Back p2 up to the last space before the wrap_length,
            if there is such a space. 

            Note that if the next character in the string (the character
            immediately after the break point) is a space, we don't need to 
            back up at all.  We'll just print up to our current location, do
            the newline, and skip to the next line. */

         if (p < charAry.length)
         {
            if (isspace(charAry[p]))
            {
               offset = 1;     /* the next character is white space.  We'll
                                  want to skip that. */
            }
            else
            {
               /* back p2 up to the last white space before the break point */

               while ((p2 > marker) && !isspace( charAry[ p2]))
               {
                  p2--;
               }

               offset = 0;
            }
         }

         /* If the line was completely filled (no place to break),
            we'll just copy the whole line out and force a break. */

         if (p2 == marker)
         {
            p2 = p-1;

            //EJMTrace.trace( EJMTrace.DEBUG, this,  "Step 3: no opportunity for break, forcing..");
         }
         else
         {
            //EJMTrace.trace( EJMTrace.DEBUG, this, "Step 3: found break at column " + p2);
         }

         if (!isspace(charAry[p2]))
         {
            /* If weren't were able to back up to a space, copy out the whole line, 
               including the break character (in this case, we'll be making the 
               string one character longer by inserting a newline). */

            result.append(inString.substring(marker, p2+1));
         }
         else
         {
            /* The break character is whitespace.  We'll copy out the characters up 
               to but not including the break character, which we will effectively 
               replace with a newline. */

            result.append(inString.substring(marker, p2));
         }


         // If we have not reached the end of the string, newline 

         if (p < charAry.length)
         {
            result.append(System.getProperty("line.separator"));

            if ( prepend != null)
            {
               result.append(prepend);
            }
         }

         p = marker = p2 + 1 + offset;
      }

      return result.toString();
   }


/**
 *
 * This method takes a string and wraps it to a line length of no more than
 * wrap_length.  
 *
 * @param inString         the string to wrap
 * @param wrap_length      the desired number of columns in each line 
 */
   private String wrap( String inString, int wrap_length)
   {
      return wrap2( inString, wrap_length, null);
   }


/**
 * Determines if a character is considered white space.
 *
 * @param   c     the character to check
 */
   private boolean isspace( char c)
   {
      return (c == '\n' || c == ' ' || c == '\t');
   }


/**
 * Output message to EJMTrace.  If a message ID is supplied use the 
 * appropriate trace level for the message:
 * <pre> Action Messages use EJMTrace.ERROR
 * Warning Messages use EJMTrace.WARN
 * Infomation messages use EJMTrace.INFO
 * all others use EJMTrace.DEBUG
 * </pre>
 */
   private void traceLog( String messageID, String message)
   {
      if (messageID == null)
      {
         //EJMTrace.trace( EJMTrace.DEBUG, "No ID", message);
         return;
      }

      if ( message.length() == 8 && messageID.startsWith("EJM") && messageID.charAt(7) == 'I') 
      {
         //EJMTrace.trace( EJMTrace.INFO, messageID, message);
         return;
      }

      if ( message.length() == 8 && messageID.startsWith("EJM") && messageID.charAt(7) == 'W') 
      {
         //EJMTrace.trace( EJMTrace.WARN, messageID, message);
         return;
      }

      if ( message.length() == 8 && messageID.startsWith("CMB") && messageID.charAt(7) == 'A') 
      {
         //EJMTrace.trace( EJMTrace.ERROR, messageID, message);
         return;
      }

      //EJMTrace.trace( EJMTrace.DEBUG, messageID, message);
   }

/**
 * Used for proper deserialization.  Sets all transient data to its normal 
 * initial state.
 *
 * @param in   the stream performing the read operation
 *
 * @exception  IOException                I/O problem
 * @exception  ClassNotFoundException     Could not find class
 */
   private void readObject( ObjectInputStream in) throws IOException, ClassNotFoundException 
   {
      in.defaultReadObject();

      fieldListener = null;
      fieldFrame = null;
      fieldFrameNotProvided = false;
      fieldButtonPanel = new JPanel();
      fieldMessageText = null;
      fieldMessageBox = null;
      fieldWidth = minWidth;
      fieldMessageID = null;
      fieldFirstButton = null;

      return;
   }

/**
 * JTextArea which overrides the is FocusTraversable method.
 * Does not accept keyboard focus if it is not editable.
 * Also overrides getColumnWidth for Chinese Locales to prevent
 * truncation of messages.
 *
 * @overrides isFocusTraversable
 * @overrides getColumnWidth
 */
class JNFTextArea extends JTextArea
{
   public JNFTextArea( String message, int rows, int columns)
   {
      super( message, rows, columns);
   }


/**
 * Text area is focus traversable only if enabled and editable.
 * No need to get keyboard focus if you can't edit.
 */
   public boolean isFocusTraversable()
   {
      return (isEditable() && isEnabled());
   }


/**
 * JTextArea picks a very bad column width for Chinese languages resulting in the
 * truncation of many messages.  This will force the column width to be wide enough
 * to display most messages.  It only affects the returned value if it is running
 * with a Chinese default locale.
 */
   protected int getColumnWidth() 
   {
      int columnWidth = super.getColumnWidth();

      if ( Locale.getDefault().getLanguage().equals( Locale.CHINESE.getLanguage())) 
      {
         if (columnWidth < 13)
         {
            columnWidth = 13;
         }
      }

      return columnWidth;
   }

}

/**
 *
 * This method takes a string and wraps it to a line length of no more than
 * wrapLength.  If prepend is not null, each resulting line will be prefixed
 * with the prepend string.  In that case, resultant line length will be no
 * more than wrapLength + prepend.length()
 *
 * @param inString         the string to wrap
 * @param wrapLength       the desired number of columns in each line 
 * @param prepend          the string to prepend to each line
 */
   private String wrap2( String inString, int wrapLength, String prepend)
   {

      if ( inString == null) 
      {
         throw new ShouldNotOccurException( "MessageBox - null message");
      }

      if ( wrapLength < 0)
      {
         throw new ShouldNotOccurException( "MessageBox - width too small: " + wrapLength);
      }

      BreakIterator boundary = BreakIterator.getLineInstance();
      String result = "";
      String current = null;

      if (prepend != null)
      {
         wrapLength = wrapLength - prepend.length();
      }
      else
         prepend = "";


      // Setup to parse text into line length (or less) units
      boundary.setText( inString);
      int start = boundary.first();
      int end = boundary.next();
      while (end != BreakIterator.DONE)
      {
         if ((end - start) > wrapLength)
         {
            end = boundary.previous();

            // No break in line - must truncate at end of line
            if (start >= end)
            {
               current = inString.substring( start, start + wrapLength);
               //EJMTrace.trace( EJMTrace.DEBUG, this,  "LL " + current.length() + " " + current);
               result = result.concat( prepend + current + "\n");
               start = start + wrapLength;
               end = boundary.next();
               continue;
            }
            else
            {
               current = inString.substring( start, end);
               //EJMTrace.trace( EJMTrace.DEBUG, this,  "NL " + current.length() + " " + current);
               start = end;
               result = result.concat( prepend + current + "\n");
            }
         }

         end = boundary.next();
      }

      current = inString.substring( start);
      //EJMTrace.trace( EJMTrace.DEBUG, this,  "EL " + current.length() + " " + current);
      result = result.concat( prepend + current);

      //EJMTrace.trace( EJMTrace.DEBUG, this,  "RS " + result.length() + " " + result);
      return result;
   }

}

